//+------------------------------------------------------------------+
//|                                               TickCompressor.mqh |
//|                               Copyright (c) 2020-2025, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| v.0.1c                       https://www.mql5.com/en/code/30791/ |
//+------------------------------------------------------------------+

#define TICK_COMPRESSOR
#define AUTODIGITS(D) (D == INT_MAX ? fmax(_Digits, 6) : D)

//MaxPrice (29-bits) = 536870911 / MathPow(10, digits)
//digits = 6, MaxPrice = 536.870911.
//digits = 5, MaxPrice = 5368.70911.
//digits = 4, MaxPrice = 53687.0911.
uint PriceToInteger(const double price, const int digits)
{
  return (uint)(price * GetPower10(digits) * (1 + DBL_EPSILON)) | (digits << 29);
}

double IntegerToPrice(const uint price)
{
  const uint digits = price >> 29;
  return NormalizeDouble((price & (UINT_MAX >> 3)) / GetPower10(digits), digits);
}

// fast integer power of 10.
double GetPower10(const int digits)
{
  const static double powers[] = {1e0, 1e1, 1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8};
  if(digits < 0 || digits > 8)
  {
    return MathPow(10, digits);
  }
  return powers[digits];
}

/*

just for reference, default tick type
struct MqlTick
{
  datetime     time;
  double       bid;
  double       ask;
  double       last;
  ulong        volume;
  long         time_msc;
  uint         flags;
  double       volume_real;
};

*/

// 2 types of Mini-Ticks: for Forex/CFDs (MqlTickBidAsk) and Exchanges (MqlTickBidAskLastVolume)

struct MqlTickBidAsk
{
  uint       bid;
  uint       ask;
  long       time_msc;
  ushort     flags;

  static MqlTickBidAsk compress(const MqlTick &tick, const int digits = INT_MAX);
  static MqlTick decompress(const MqlTickBidAsk &tick);
};

struct MqlTickBidAskLastVolume
{
  uint       bid;
  uint       ask;
  uint       last;
  ulong      volume;
  long       time_msc;
  ushort     flags;

  static MqlTickBidAskLastVolume compress(const MqlTick &tick, const int digits = INT_MAX);
  static MqlTick decompress(const MqlTickBidAskLastVolume &tick);
};


// compression/decompression static class wrapper
class TickCompressor
{
  public:
    // dummy compressions/decompressions
    // to keep dependent codes unchanged for native and mini ticks

    static void compress(const MqlTick &tick, MqlTick &result)
    {
      result = tick;
    }

    static void decompress(const MqlTick &tick, MqlTick &result)
    {
      result = tick;
    }

    static void compress(MqlTick &ticks[], MqlTick &result[])
    {
      ArraySwap(ticks, result);
    }

    static void decompress(MqlTick &ticks[], MqlTick &result[])
    {
      ArraySwap(ticks, result);
    }

    // Forex/CFDs
    static void compress(const MqlTick &tick, MqlTickBidAsk &result, const int digits = INT_MAX)
    {
      result.bid = PriceToInteger(tick.bid, AUTODIGITS(digits));
      result.ask = PriceToInteger(tick.ask, AUTODIGITS(digits));
      result.time_msc = tick.time_msc;
      result.flags = (ushort)tick.flags;
    }

    static void decompress(const MqlTickBidAsk &tick, MqlTick &result)
    {
      result.bid = IntegerToPrice(tick.bid);
      result.ask = IntegerToPrice(tick.ask);
      result.last = 0;
      result.time_msc = tick.time_msc;
      result.time = (datetime)(tick.time_msc / 1000);
      result.flags = tick.flags;
      result.volume = 0;
      result.volume_real = 0;
    }


    // Exchanges
    static void compress(const MqlTick &tick, MqlTickBidAskLastVolume &result, const int digits = INT_MAX)
    {
      result.bid = PriceToInteger(tick.bid, AUTODIGITS(digits));
      result.ask = PriceToInteger(tick.ask, AUTODIGITS(digits));
      result.last = PriceToInteger(tick.last, AUTODIGITS(digits));
      result.volume = tick.volume;
      result.time_msc = tick.time_msc;
      result.flags = (ushort)tick.flags;
    }

    static void decompress(const MqlTickBidAskLastVolume &tick, MqlTick &result)
    {
      result.bid = IntegerToPrice(tick.bid);
      result.ask = IntegerToPrice(tick.ask);
      result.last = IntegerToPrice(tick.last);
      result.volume = tick.volume;
      result.time_msc = tick.time_msc;
      result.time = (datetime)(tick.time_msc / 1000);
      result.flags = tick.flags;
      result.volume_real = (double)tick.volume;
    }

    // arrays
    template<typename D>
    static void compress(const MqlTick &ticks[], D &result[], const int digits = INT_MAX)
    {
      const int n = ArraySize(ticks);
      ArrayResize(result, n);
      for(int i = 0; i < n; i++)
      {
        compress(ticks[i], result[i], digits);
      }
    }

    template<typename S>
    static void decompress(const S &ticks[], MqlTick &result[])
    {
      const int n = ArraySize(ticks);
      ArrayResize(result, n);
      for(int i = 0; i < n; i++)
      {
        decompress(ticks[i], result[i]);
      }
    }

};

static MqlTickBidAsk MqlTickBidAsk::compress(const MqlTick &tick, const int digits = INT_MAX)
{
  MqlTickBidAsk result = {0};
  TickCompressor::compress(tick, result, digits);
  return result;
}

static MqlTick MqlTickBidAsk::decompress(const MqlTickBidAsk &tick)
{
  MqlTick result = {0};
  TickCompressor::decompress(tick, result);
  return result;
}

static MqlTickBidAskLastVolume MqlTickBidAskLastVolume::compress(const MqlTick &tick, const int digits = INT_MAX)
{
  MqlTickBidAskLastVolume result;
  TickCompressor::compress(tick, result, digits);
  return result;
}

static MqlTick MqlTickBidAskLastVolume::decompress(const MqlTickBidAskLastVolume &tick)
{
  MqlTick result;
  TickCompressor::decompress(tick, result);
  return result;
}
